<?php
/**
 * Horde YAML package
 *
 * This package is heavily inspired by the Spyc PHP YAML
 * implementation (http://spyc.sourceforge.net/), and portions are
 * copyright 2005-2006 Chris Wanstrath.
 *
 * @package Horde_Yaml
 */

/**
 * Horde YAML parser.
 *
 * This class can be used to read a YAML file and convert its contents
 * into a PHP array. The native PHP parser supports a limited
 * subsection of the YAML spec, but if the syck extension is present,
 * that will be used for parsing.
 *
 * Usage:
 * <code>
 *   $array = Horde_Yaml::load($file);
 * </code>
 *
 * @package Horde_Yaml
 */
class Horde_Yaml {

    /**
     * Load YAML into a PHP array.
     *
     * The load method, when supplied with a YAML stream (string or file),
     * will do its best to convert YAML in a file into a PHP array.  Pretty
     * simple.
     *  Usage:
     *  <code>
     *   $array = Horde_Yaml::YAMLLoad('lucky.yaml');
     *   print_r($array);
     *  </code>
     * @return array
     * @param string $input Path of YAML file or string containing YAML
     */
    public static function load($yaml)
    {
        if (function_exists('syck_load')) {
            $array = syck_load($yaml);
            return is_array($array) ? $array : array();
        }

        if (strpos($yaml, "\r") !== false) {
            $yaml = str_replace(array("\r\n", "\r"), array("\n", "\n"), $yaml);
        }
        $lines = explode("\n", $yaml);
        $loader = new Horde_Yaml_Loader;

        while (list(,$line) = each($lines)) {
            $loader->parse($line);
        }

        return $loader->toArray();
    }

    /**
     */
    public static function loadFile($filename)
    {
        $oldDisplayErrors = ini_set('display_errors', 0);
        $oldTrackErrors   = ini_set('track_errors',   1);

        $stream = fopen($filename, 'rb');

        ini_set('display_errors', $oldDisplayErrors);
        ini_set('track_errors',   $oldTrackErrors);

        if (!$stream) {
            throw new Horde_Yaml_Exception($php_errormsg);
        }

        return self::loadStream($stream);
    }

    /**
     */
    public static function loadStream($stream)
    {
        if (function_exists('syck_load')) {
            $array = syck_load(stream_get_contents($stream));
            return is_array($array) ? $array : array();
        }

        $loader = new Horde_Yaml_Loader;
        while (!feof($stream)) {
            $loader->parse(stream_get_line($stream, 100000, "\n"));
        }

        return $loader->toArray();
    }

    /**
     * Dump a PHP array to YAML.
     *
     * The dump method, when supplied with an array, will do its best
     * to convert the array into friendly YAML.
     *
     * @return string
     * @param array $array PHP array
     * @param integer $indent Number of spaces to indent lines. Defaults to 2.
     * @param integer $wordwrap Cut-off for wrapping long lines. Defaults to 40; 0 means no wrap.
     */
    public static function dump($array, $indent = false, $wordwrap = false)
    {
        // TODO: variants for dump to string vs. dump to stream
        $dumper = new Horde_Yaml_Dumper;
        return $dumper->dump($array, $indent, $wordwrap);
    }

}
