<?php
/**
 * Horde YAML package
 *
 * This package is heavily inspired by the Spyc PHP YAML
 * implementation (http://spyc.sourceforge.net/), and portions are
 * copyright 2005-2006 Chris Wanstrath.
 *
 * @package Horde_Yaml
 */

/**
 * Dump PHP data structures to YAML. 
 *
 * @package Horde_Yaml
 */
class Horde_Yaml_Dumper {
    
    /**
     * Number of spaces to indent
     * @var integer
     */
    protected $_indent;
    
    /**
     * Perform wordwrap?
     * @var boolean
     */
    protected $_wordWrap;

    /**
     * Dump PHP array to YAML
     *
     * The dump method, when supplied with an array, will do its best
     * to convert the array into friendly YAML.  Pretty simple.  Feel free to
     * save the returned string as tasteful.yaml and pass it around.
     *
     * Oh, and you can decide how big the indent is and what the wordwrap
     * for folding is.  Pretty cool -- just pass in 'false' for either if
     * you want to use the default.
     *
     * Indent's default is 2 spaces, wordwrap's default is 40 characters.  And
     * you can turn off wordwrap by passing in 0.
     *
     * @param  array    $array     PHP array
     * @param  integer  $indent    Pass in false to use the default, which is 2
     * @param  integer  $wordwrap  Pass in 0 for no wordwrap, false for default (40)
     * @return string
     */
    public function dump($array, $indent = false, $wordwrap = false)
    {
        // Dumps to some very clean YAML. We'll have to add some more
        // features and options soon. And better support for folding.

        // New features and options.
        if ($indent === false || !is_numeric($indent)) {
            $this->_indent = 2;
        } else {
            $this->_indent = $indent;
        }

        if ($wordwrap === false || !is_numeric($wordwrap)) {
            $this->_wordWrap = 40;
        } else {
            $this->_wordWrap = $wordwrap;
        }

        // New YAML document
        $string = "---\n";

        // Start at the base of the array and move through it.
        foreach ($array as $key => $value) {
            $string .= $this->_yamlize($key, $value, 0);
        }
        return $string;
    }

    /**
     * Attempts to convert a key / value array item to YAML
     *
     * @param  string        $key     The name of the key
     * @param  string|array  $value   The value of the item
     * @param  integer       $indent  The indent of the current node
     * @return string
     */
    protected function _yamlize($key, $value, $indent)
    {
        if (is_array($value)) {
            // It has children.  Make it the right kind of item.
            $string = $this->_dumpNode($key, null, $indent);

            // Add the indent.
            $indent += $this->_indent;

            // Yamlize the array.
            $string .= $this->_yamlizeArray($value, $indent);
        } elseif (!is_array($value)) {
            // No children.
            $string = $this->_dumpNode($key, $value, $indent);
        }

        return $string;
    }

    /**
     * Attempts to convert an array to YAML
     *
     * @param  array    $array The array you want to convert
     * @param  integer  $indent The indent of the current level
     * @return string
     */
    protected function _yamlizeArray($array, $indent)
    {
        if (!is_array($array)) {
            return false;
        }

        $string = '';
        foreach ($array as $key => $value) {
            $string .= $this->_yamlize($key, $value, $indent);
        }
        return $string;
    }

    /**
     * Returns YAML from a key and a value
     *
     * @param  string   $key     The name of the key
     * @param  string   $value   The value of the item
     * @param  integer  $indent  The indent of the current node
     * @return string
     */
    protected function _dumpNode($key, $value, $indent)
    {
        // Do some folding here, for blocks.
        if (strpos($value, "\n") !== false
            || strpos($value, ': ') !== false
            || strpos($value, '- ') !== false) {
            $value = $this->_doLiteralBlock($value, $indent);
        } else {
            $value = $this->_fold($value, $indent);
        }

        if (is_bool($value)) {
            $value = ($value) ? 'true' : 'false';
        }

        $spaces = str_repeat(' ', $indent);

        if (is_int($key)) {
            // It's a sequence.
            $string = $spaces . '- ' . $value . "\n";
        } else {
            // It's mapped.
            $string = $spaces . $key . ': ' . $value . "\n";
        }

        return $string;
    }

    /**
     * Creates a literal block for dumping
     *
     * @param  string   $value
     * @param  integer  $indent  The value of the indent.
     * @return string
     */
    protected function _doLiteralBlock($value, $indent)
    {
        $exploded = explode("\n", $value);
        $newValue = '|';
        $indent += $this->_indent;
        $spaces = str_repeat(' ', $indent);
        foreach ($exploded as $line) {
            $newValue .= "\n" . $spaces . trim($line);
        }
        return $newValue;
    }

    /**
     * Folds a string of text, if necessary
     *
     * @param   $value   The string you wish to fold
     * @return  string
     */
    protected function _fold($value, $indent)
    {
        // Don't do anything if wordwrap is set to 0
        if ($this->_wordWrap === 0) {
            return $value;
        }

        if (strlen($value) > $this->_wordWrap) {
            $indent += $this->_indent;
            $indent = str_repeat(' ', $indent);
            $wrapped = wordwrap($value, $this->_wordWrap, "\n$indent");
            $value = ">\n" . $indent . $wrapped;
        }

        return $value;
    }

}
