<?php
/**
 * @package SyncML
 *
 * $Horde: framework/SyncML/SyncML/Device/Sync4j.php,v 1.35 2007/06/27 17:23:03 jan Exp $
 */

/** Horde_Date */
require_once 'Horde/Date.php';

/** Horde_iCalendar */
require_once 'Horde/iCalendar.php';

/**
 * Sync4j (www.sync4j.org)
 *
 * The Sync4J outlook converter uses its native SIF format for data
 * exchange. Conversion to text/vcalendar etc. is done by
 * SifConverter.php The connector seems not support DevInf
 * information, so SyncML_Device can only detect it by the decice ID:
 * so in the connector configuration the device ID must be set to
 * 'sc-pim-<type>' which should be the default anyhow.
 *
 * Copyright 2005-2007 Karsten Fourmont <karsten@horde.org>
 *
 * See the enclosed file COPYING for license information (LGPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/lgpl.html.
 *
 * @author  Karsten Fourmont <karsten@horde.org>
 * @package SyncML
 */
class SyncML_Device_sync4j extends SyncML_Device {

    function getPreferredContentTypeClient($serverSyncURI, $sourceSyncURI)
    {
        if (strpos(strtolower($serverSyncURI), 'note') !== false
            || strpos(strtolower($serverSyncURI), 'memo') !== false) {
            return 'text/x-vnote';
        }
        return parent::getPreferredContentTypeClient($serverSyncURI, $sourceSyncURI);
    }

    /**
     * Convert the content.
     *
     * Currently strips uid (primary key) information as client and
     * server might use different ones.
     *
     * Charset conversions might be added here too.
     */
    function convertClient2Server($content, $contentType)
    {
        list($content, $contentType) =
            parent::convertClient2Server(base64_decode($content), $contentType);

        switch ($contentType) {
        case 'text/x-s4j-sifn' :
            $content = SyncML_Device_sync4j::sif2vnote($content);
            $contentType = 'text/x-vnote';
            break;

        case 'text/x-s4j-sifc' :
            $content = SyncML_Device_sync4j::sif2vcard($content);
            $contentType = 'text/x-vcard';
            break;

        case 'text/x-s4j-sife' :
            $content = SyncML_Device_sync4j::sif2vevent($content);
            $contentType = 'text/x-vevent';
            break;

        case 'text/x-s4j-sift' :
            $content = SyncML_Device_sync4j::sif2vtodo($content);
            $contentType = 'text/x-vtodo';
            break;
        }

        $l = "\ninput converted for server ($contentType):\n" . $content . "\n";
        $GLOBALS['backend']->logFile(SYNCML_LOGFILE_DATA, $l);

        return array($content, $contentType);
    }

    /**
     * Converts the content from the backend to a format suitable for the
     * client device.
     *
     * Strips the uid (primary key) information as client and server might use
     * different ones.
     *
     * @param string $content       The content to convert
     * @param string $contentType   The contentType of content as returned from
     *                              the backend
     * @return array                array($newcontent, $newcontentType, $newEncodingType):
     *                              the converted content and the
     *                              (possibly changed) new ContentType and EncodingType
     *                              (like b64 as used by Funambol/Sync4J).
     */
    function convertServer2Client($content, $contentType)
    {
        global $backend;

        list($content, $contentType, $encodingType) =
            parent::convertServer2Client($content, $contentType);

        switch ($contentType) {
        case 'text/calendar' :
        case 'text/x-icalendar' :
        case 'text/x-vcalendar' :
        case 'text/x-vevent' :
            $content = SyncML_Device_sync4j::vevent2sif($content);
            $content = base64_encode($content);
            $contentType = 'text/x-s4j-sife';
            break;

        case 'text/x-vtodo' :
            $content = SyncML_Device_sync4j::vtodo2sif($content);
            $content = base64_encode($content);
            $contentType = 'text/x-s4j-sift';
            break;

        case 'text/x-vcard' :
            $content = SyncML_Device_sync4j::vcard2sif($content);
            $content = base64_encode($content);
            $contentType = 'text/x-s4j-sifc';
            break;

        case 'text/x-vnote':
        case 'text/plain':
            $content = SyncML_Device_sync4j::vnote2sif($content);
            $content = base64_encode($content);
            $contentType = 'text/x-s4j-sifn';
            break;
        }

        $l = "\noutput converted for client ($contentType):\n" . base64_decode($content) . "\n";
        $GLOBALS['backend']->logFile(SYNCML_LOGFILE_DATA, $l);

        return array($content, $contentType, 'b64');
    }

    /**
     * Decodes a sif xml string to an associative array.
     *
     * Quick hack to convert from text/vcard and text/vcalendar to
     * Sync4J's proprietery sif datatypes and vice versa.  For details
     * about the sif format see the appendix of the developer guide on
     * www.sync4j.org.
     *
     * @access private
     *
     * @param string $sif  A sif string like <k1>v1&gt;</k1><k2>v2</k2>
     *
     * @return array  Assoc array in utf8 like array ('k1' => 'v1>', 'k2' => 'v2');
     */
    function sif2array($sif)
    {
        $r = array();
        if (preg_match_all('/<([^>]*)>([^<]*)<\/\1>/si', $sif, $matches, PREG_SET_ORDER)) {
            foreach ($matches as $match) {
                $r[$match[1]] = html_entity_decode($match[2]);
            }
        }
        return $r;
    }

    /**
     * Encodes an assoc. array to sif xml
     *
     * Quick hack to convert from text/vcard and text/vcalendar to
     * Sync4J's proprietery sif datatypes and vice versa.  For details
     * about the sif format see the appendix of the developer guide on
     * www.sync4j.org.
     *
     * @access private
     *
     * @param array $array  An assoc array.
     *
     * @return string  The resulting XML string.
     */
    function array2sif($array, $pre = '', $post = '')
    {
        $search = array('<', '>');
        $replace = array('&lt;', '&gt;');

        // Sorting the array is not strictly necessary, but it ensures
        // reproducible results when doing tests and comparing the output
        // against pre-recorded data.
        ksort($array);
        $r = $pre;
        foreach ($array as $key => $value) {
            if (is_a($value, 'PEAR_Error')) {
                continue;
            }
            if (is_array($value)) {
                $value = $value[0];
            }
            $r .= '<' . $key . '>' .
                str_replace($search, $replace, $value) .
                '</' . $key . '>';
        }

        return $r . $post;
    }

    function sif2vnote($sif)
    {
        $a = SyncML_Device_sync4j::sif2array($sif);

        $iCal = &new Horde_iCalendar();
        $iCal->setAttribute('VERSION', '1.1');
        $iCal->setAttribute('PRODID', '-//The Horde Project//Mnemo//EN');
        $iCal->setAttribute('METHOD', 'PUBLISH');

        $vnote = &Horde_iCalendar::newComponent('vnote', $iCal);
        $vnote->setAttribute('BODY', isset($a['Body']) ? $a['Body'] : '');
        if (isset($a['Subject'])) {
            $vnote->setAttribute('SUMMARY', $a['Subject']);
        }
        if (isset($a['Categories'])) {
            $vnote->setAttribute('CATEGORIES', $a['Categories']);
        }

        return $vnote->exportvCalendar();
    }

    function sif2vcard($sif)
    {
        $a = SyncML_Device_sync4j::sif2array($sif);

        $iCal = &new Horde_iCalendar();
        $iCal->setAttribute('VERSION', '2.1');
        $iCal->setAttribute('PRODID', '-//The Horde Project//Turba//EN');
        $iCal->setAttribute('METHOD', 'PUBLISH');

        $vcard = &Horde_iCalendar::newComponent('vcard', $iCal);

        $vcard->setAttribute('FN', $a['FileAs']);
        $vcard->setAttribute('NICKNAME', $a['NickName']);
        $vcard->setAttribute('TEL', $a['HomeTelephoneNumber'],
                             array('TYPE'=>'HOME'));
        $vcard->setAttribute('TEL', $a['BusinessTelephoneNumber'],
                             array('TYPE'=>'WORK'));
        $vcard->setAttribute('TEL', $a['MobileTelephoneNumber'],
                             array('TYPE'=>'CELL'));
        $vcard->setAttribute('TEL', $a['BusinessFaxNumber'],
                             array('TYPE'=>'FAX'));
        $vcard->setAttribute('EMAIL', $a['Email1Address']);
        $vcard->setAttribute('TITLE', $a['JobTitle']);
        $vcard->setAttribute('ORG', str_replace(array('\r\n', ';', ':'), array('\\n', '\\;', '\\:'), $a['CompanyName'])
                             . ';' . str_replace(array('\r\n', ';', ':'), array('\\n', '\\;', '\\:'), $a['Department']));
        $vcard->setAttribute('NOTE', $a['Body']);
        $vcard->setAttribute('URL', $a['WebPage']);
        $vcard->setAttribute('BDAY', $a['Birthday']);

        $v = array(
            VCARD_N_FAMILY      => $a['LastName'],
            VCARD_N_GIVEN       => $a['FirstName'],
            VCARD_N_ADDL        => $a['MiddleName'],
            VCARD_N_PREFIX      => $a['Title'],
            VCARD_N_SUFFIX      => $a['Suffix']
        );
        $vcard->setAttribute('N', implode(';', $v), array(), false, $v);

        $v = array(
            VCARD_ADR_POB           => $a['HomeAddressPostOfficeBox'],
            VCARD_ADR_EXTEND        => '',
            VCARD_ADR_STREET        => $a['HomeAddressStreet'],
            VCARD_ADR_LOCALITY      => $a['HomeAddressCity'],
            VCARD_ADR_REGION        => $a['HomeAddressState'],
            VCARD_ADR_POSTCODE      => $a['HomeAddressPostalCode'],
            VCARD_ADR_COUNTRY       => $a['HomeAddressCountry'],
        );
        $vcard->setAttribute('ADR', implode(';', $v), array('TYPE' => 'HOME'), true, $v);

        $v = array(
            VCARD_ADR_POB           => $a['BusinessAddressPostOfficeBox'],
            VCARD_ADR_EXTEND        => '',
            VCARD_ADR_STREET        => $a['BusinessAddressStreet'],
            VCARD_ADR_LOCALITY      => $a['BusinessAddressCity'],
            VCARD_ADR_REGION        => $a['BusinessAddressState'],
            VCARD_ADR_POSTCODE      => $a['BusinessAddressPostalCode'],
            VCARD_ADR_COUNTRY       => $a['BusinessAddressCountry'],
        );
        $vcard->setAttribute('BDAY', isset($a['Birthday']) ? $a['Birthday'] : '');

        $vcard->setAttribute('X-SYNCJE-ANNIVERSARY', isset($a['Anniversary']) ? $a['Anniversary'] : '');
        $vcard->setAttribute('X-SYNCJE-SPOUSE', isset($a['Spouse']) ? $a['Spouse'] : '');
        $vcard->setAttribute('X-SYNCJE-CHILD', isset($a['Children']) ? $a['Children'] : '');

        $vcard->setAttribute('ADR', implode(';', $v), array('TYPE' => 'WORK'), true, $v);
        $vcard->setAttribute('CATEGORIES', isset($a['Categories']) ? $a['Categories'] : '');

        return $vcard->exportvCalendar();
    }

    function sif2vevent($sif)
    {
        $a = SyncML_Device_sync4j::sif2array($sif);

        $iCal = &new Horde_iCalendar();
        $iCal->setAttribute('VERSION', '1.0');
        $iCal->setAttribute('PRODID', '-//The Horde Project//Kronolith//EN');
        $iCal->setAttribute('METHOD', 'PUBLISH');

        $vEvent = &Horde_iCalendar::newComponent('vevent', $iCal);
        $vEvent->setAttribute('DTSTAMP', time());

        $map = array('Subject' => 'SUMMARY',
                     'Body' => 'DESCRIPTION',
                     'Categories' => 'CATEGORIES',
                     'Location' => 'LOCATION');
        foreach ($map as $source => $target) {
            if (!empty($a[$source])) {
                $vEvent->setAttribute($target, $a[$source]);
            }
        }

        if ($a['AllDayEvent'] == 1) {
            $startTime = $iCal->_parseDateTime($a['Start']);
            $vEvent->setAttribute('DTSTART',
                                  array('year' => date('Y', $startTime),
                                        'month' => date('m', $startTime),
                                        'mday' => date('d', $startTime)),
                                  array('VALUE' => 'DATE'));
            $t = $iCal->_parseDateTime($a['End']);
            $vEvent->setAttribute('DTEND',
                                  array('year' => date('Y', $t),
                                        'month' => date('m', $t),
                                        'mday' => date('d', $t)),
                                  array('VALUE' => 'DATE'));
        } else {
            $startTime = $iCal->_parseDateTime($a['Start']);
            $vEvent->setAttribute('DTSTART', $startTime);
            $vEvent->setAttribute('DTEND',
                                  $iCal->_parseDateTime($a['End']));
        }

        if (isset($a['IsRecurring']) && $a['IsRecurring'] == 1) {
            $interval = '';
            switch ($a['RecurrenceType']) {
            case 0:
                /* olDaily */
                if (!empty($a['DayOfWeekMask'])) {
                    $freq = 'WEEKLY';
                    $interval = ';INTERVAL=1';
                } else {
                    $freq = 'DAILY';
                    $interval = ';INTERVAL=' . $a['Interval'];
                }
                break;
            case 1:
                /* olWeekly */
                $freq = 'WEEKLY';
                $interval = ';INTERVAL=' . $a['Interval'];
                break;
            case 2:
                /* olMonthly */
                $freq = 'MONTHLY';
                $interval = ';INTERVAL=' . $a['Interval'];
                break;
            case 3:
                /* olMonthNth */
                $freq = 'MONTHLY';
                $interval = ';INTERVAL=' . $a['Interval'];
                break;
            case 5:
                /* olYearly */
                $freq = 'YEARLY';
                break;
            case 6:
                /* olYearNth */
                $freq = 'YEARLY';
                break;
            }
            $rrule = 'FREQ=' . $freq;
            if (!isset($a['NoEndDate']) || $a['NoEndDate'] != 1) {
                $rrule .= ';UNTIL=' . $a['PatternEndDate'];
            }
            $rrule .= $interval;
            if (!empty($a['DayOfMonth'])) {
                $rrule .= ';BYMONTHDAY=' . $a['DayOfMonth'];
            }
            if (!empty($a['MonthOfYear'])) {
                $rrule .= ';BYMONTH=' . $a['MonthOfYear'];
            }
            if (!empty($a['DayOfWeekMask'])) {
                $rrule .= ';BYDAY=';
                $icaldays = array('SU', 'MO', 'TU', 'WE', 'TH', 'FR', 'SA');
                for ($i = $flag = 0; $i <= 7 ; ++$i) {
                    if (pow(2, $i) & $a['DayOfWeekMask']) {
                        if ($flag) {
                            $rrule .= ',';
                        }
                        $rrule .= $icaldays[$i];
                        $flag = true;
                    }
                }
            }
            $vEvent->setAttribute('RRULE', $rrule);
        }

        if (isset($a['ReminderSet']) && $a['ReminderSet'] == 1) {
            $vEvent->setAttribute('AALARM', $startTime - $a['ReminderMinutesBeforeStart'] * 60);
        }

        if (isset($a['BusyStatus'])) {
            switch ($a['BusyStatus']) {
            case 0:
                /* olFree - FREE is not a iCalendar standard value. */
                $vEvent->setAttribute('STATUS', 'FREE');
                break;
            case 1:
                /* olTentative */
                $vEvent->setAttribute('STATUS', 'TENTATIVE');
                break;
            case 2:
                /* olBusy */
            case 3:
                /* olOutOfOffice */
                $vEvent->setAttribute('STATUS', 'CONFIRMED');
                break;
            }
        }

        if (isset($a['Sensitivity'])) {
            switch ($a['Sensitivity']) {
            case 0:
                /* olNormal - FREE is not a iCalendar standard value. */
                $vEvent->setAttribute('CLASS', 'PUBLIC');
                break;
            case 1:
                /* olPersonal */
            case 2:
                /* olPrivate */
                $vEvent->setAttribute('CLASS', 'PRIVATE');
                break;
            case 3:
                /* olConfidential */
                $vEvent->setAttribute('CLASS', 'CONFIDENTIAL');
                break;
            }
        }

        return $vEvent->exportvCalendar();
    }

    function sif2vtodo($sif)
    {
        $a = SyncML_Device_sync4j::sif2array($sif);

        $iCal = &new Horde_iCalendar();
        $iCal->setAttribute('VERSION', '1.0');
        $iCal->setAttribute('PRODID', '-//The Horde Project//Nag//EN');
        $iCal->setAttribute('METHOD', 'PUBLISH');

        $vtodo = &Horde_iCalendar::newComponent('vtodo', $iCal);

        $vtodo->setAttribute('SUMMARY', $a['Subject']);
        $vtodo->setAttribute('DESCRIPTION', $a['Body']);
        if ($a['Importance'] == 0) {
            $vtodo->setAttribute('PRIORITY', 5);
        } elseif ($a['Importance'] == 2) {
            $vtodo->setAttribute('PRIORITY', 1);
        } else {
            $vtodo->setAttribute('PRIORITY', 3);
        }
        if (!empty($a['DueDate']) && $a['DueDate'] != '45001231T230000Z') {
            $vtodo->setAttribute('DUE', $iCal->_parseDateTime($a['DueDate']));
        }
        $vtodo->setAttribute('COMPLETED', $a['Complete'] == 'True' ? 1 : 0);
        $vtodo->setAttribute('CATEGORIES', isset($a['Categories']) ? $a['Categories'] : '');

        return $vtodo->exportvCalendar();
    }

    function vnote2sif($vnote)
    {
        if (strpos($vnote, 'BEGIN:') === false) {
            // handle plain text:
            $a = array('Body' => $vnote);
        } else {
            $iCal = new Horde_iCalendar();
            if (!$iCal->parsevCalendar($vnote)) {
                // @TODO: NEVER use die() in a library.
                die("There was an error importing the data.");
            }

            $components = $iCal->getComponents();
            if (!is_Array($components) || count($components) == 0) {
                $a = array('Body' =>
                           'Conversion error in Sync4j.php: no data found!');
            } else {
                $a = array('Body' => $components[0]->getAttribute('BODY'),
                       'Categories' => $components[0]->getAttribute('CATEGORIES'));
                $sum = $components[0]->getAttribute('SUMMARY');
                if (!is_a($sum, 'PEAR_Error')) {
                    $a['Subject'] = $sum;
                }
            }
        }

        return SyncML_Device_sync4j::array2sif(
            $a,
            '<note>',
            '<Color></Color><Height></Height><Width></Width><Left></Left><Top></Top>'
            . '</note>');
    }

    function vcard2sif($vcard)
    {
        $iCal = &new Horde_iCalendar();
        if (!$iCal->parsevCalendar($vcard)) {
            // @TODO: NEVER use die() in a library.
            die("There was an error importing the data.");
        }

        $components = $iCal->getComponents();

        switch (count($components)) {
        case 0:
            // @TODO: NEVER use die() in a library.
            die("No data was found.");

        case 1:
            $content = $components[0];
            break;

        default:
            // @TODO: NEVER use die() in a library.
            die("Multiple components found; only one is supported.");
        }

        // from here on, the code is taken from
        // Turba_Driver::toHash, v 1.65 2005/03/12
        // and modified for the Sync4J attribute names.
        $attr = $content->getAllAttributes();
        foreach ($attr as $item) {
            switch ($item['name']) {
            case 'FN':
                $hash['FileAs'] = $item['value'];
                break;

            case 'N':
                $name = $item['values'];
                $hash['LastName'] = $name[VCARD_N_FAMILY];
                $hash['FirstName'] = $name[VCARD_N_GIVEN];
                $hash['MiddleName'] = $name[VCARD_N_ADDL];
                $hash['Title'] = $name[VCARD_N_PREFIX];
                $hash['Suffix'] = $name[VCARD_N_SUFFIX];
                break;

            case 'NICKNAME':
                $hash['NickName'] = $item['value'];
                break;

            // For vCard 3.0.
            case 'ADR':
                if (isset($item['params']['TYPE'])) {
                    if (!is_array($item['params']['TYPE'])) {
                        $item['params']['TYPE'] = array($item['params']['TYPE']);
                    }
                } else {
                    $item['params']['TYPE'] = array();
                    if (isset($item['params']['WORK'])) {
                        $item['params']['TYPE'][] = 'WORK';
                    }
                    if (isset($item['params']['HOME'])) {
                        $item['params']['TYPE'][] = 'HOME';
                    }
                }

                $address = $item['values'];
                foreach ($item['params']['TYPE'] as $adr) {
                    switch (String::upper($adr)) {
                    case 'HOME':
                        $prefix = 'HomeAddress';
                        break;

                    case 'WORK':
                        $prefix = 'BusinessAddress';
                        break;

                    default:
                        $prefix = 'HomeAddress';
                    }

                    if ($prefix) {
                        $hash[$prefix . 'Street'] = isset($address[VCARD_ADR_STREET]) ? $address[VCARD_ADR_STREET] : null;
                        $hash[$prefix . 'City'] = isset($address[VCARD_ADR_LOCALITY]) ? $address[VCARD_ADR_LOCALITY] : null;
                        $hash[$prefix . 'State'] = isset($address[VCARD_ADR_REGION]) ? $address[VCARD_ADR_REGION] : null;
                        $hash[$prefix . 'PostalCode'] = isset($address[VCARD_ADR_POSTCODE]) ? $address[VCARD_ADR_POSTCODE] : null;
                        $hash[$prefix . 'Country'] = isset($address[VCARD_ADR_COUNTRY]) ? $address[VCARD_ADR_COUNTRY] : null;
                        $hash[$prefix . 'PostOfficeBox'] = isset($address[VCARD_ADR_POB]) ? $address[VCARD_ADR_POB] : null;
                    }
                }
                break;

            case 'TEL':
                if (isset($item['params']['FAX'])) {
                    $hash['BusinessFaxNumber'] = $item['value'];
                } elseif (isset($item['params']['TYPE'])) {
                    if (!is_array($item['params']['TYPE'])) {
                        $item['params']['TYPE'] = array($item['params']['TYPE']);
                    }
                    // For vCard 3.0.
                    foreach ($item['params']['TYPE'] as $tel) {
                        if (String::upper($tel) == 'WORK') {
                            $hash['BusinessTelephoneNumber'] = $item['value'];
                        } elseif (String::upper($tel) == 'HOME') {
                            $hash['HomeTelephoneNumber'] = $item['value'];
                        } elseif (String::upper($tel) == 'CELL') {
                            $hash['MobileTelephoneNumber'] = $item['value'];
                        } elseif (String::upper($tel) == 'FAX') {
                            $hash['BusinessFaxNumber'] = $item['value'];
                        }
                    }
                } else {
                    if (isset($item['params']['HOME'])) {
                        $hash['HomeTelephoneNumber'] = $item['value'];
                    } elseif (isset($item['params']['WORK'])) {
                        $hash['BusinessTelephoneNumber'] = $item['value'];
                    } elseif (isset($item['params']['CELL'])) {
                        $hash['MobileTelephoneNumber'] = $item['value'];
                    } else {
                        $hash['HomeTelephoneNumber'] = $item['value'];
                    }
                }
                break;

            case 'EMAIL':
                if (isset($item['params']['PREF']) || !isset($hash['email'])) {
                    $hash['Email1Address'] = Horde_iCalendar_vcard::getBareEmail($item['value']);
                    $hash['Email1AddressType'] = 'SMTP';
                }
                break;

            case 'TITLE':
                $hash['JobTitle'] = $item['value'];
                break;

            case 'ORG':
                $values = preg_split('/(?<!\\\\);/', trim($item['value']));
                $hash['CompanyName'] = $values[0];
                $hash['Department'] = isset($values[1]) ? $values[1] : '';
                break;

            case 'NOTE':
                $hash['Body'] = $item['value'];
                break;

            case 'URL':
                $hash['WebPage'] = $item['value'];
                break;

            case 'BDAY':
                $hash['Birthday'] = $item['value'];
                break;

            case 'X-SYNCJE-ANNIVERSARY':
                $hash['Anniversary'] = $item['value'];
                break;

            case 'X-SYNCJE-SPOUSE':
                $hash['Spouse'] = $item['value'];
                break;

            case 'X-SYNCJE-CHILD':
                $hash['Children'] = $item['value'];
                break;

            case 'BDAY':
                if (is_array($item['value'])) {
                    $hash['Birthday'] = sprintf('%04d-%02d-%02d',
                                                $item['value']['year'],
                                                $item['value']['month'],
                                                $item['value']['mday']);
                }
                break;

            case 'CATEGORIES':
                $hash['Categories'] = $item['value'];
                break;
            }
        }

        return SyncML_Device_sync4j::array2sif(
            $hash,
            '<?xml version="1.0"?><contact>',
            '</contact>');
    }

    function vevent2sif($vcard)
    {
        /* some special handling for all day vevents that are not passed
         * as TYPE=DATE (TYPE=DATE does not exist for VCALENDAR1.0)
         */
        if (preg_match('/(\r\n|\r|\n)DTSTART:.*T000000(\r\n|\r|\n)/', $vcard)) {
            if (preg_match('/(\r\n|\r|\n)DTEND:(\d\d\d\d)(\d\d)(\d\d)T235959(\r\n|\r|\n)/', $vcard, $m)) {
                $vcard = preg_replace('/(\r\n|\r|\n)DTSTART:(.*)T000000(\r\n|\r|\n)/',
                                      "$1DTSTART;VALUE=DATE:$2$3", $vcard);
                $vcard = preg_replace('/(\r\n|\r|\n)DTEND:(.*)T235959(\r\n|\r|\n)/',
                                      "$1DTEND;VALUE=DATE:$2$3", $vcard);
            }
            // @TODO: else: handle case with DTEND= T240000
        }
        $iCal = new Horde_iCalendar();
        if (!$iCal->parsevCalendar($vcard)) {
            // @TODO: NEVER use die() in a library.
            die("There was an error importing the data.");
        }

        $components = $iCal->getComponents();

        switch (count($components)) {
        case 0:
            // @TODO: NEVER use die() in a library.
            die("No data was found.");

        case 1:
            $content = $components[0];
            break;

        default:
            // @TODO: NEVER use die() in a library.
            die("Multiple components found; only one is supported.");
        }

        $hash = array('ReminderSet' => 0,
                      'IsRecurring' => 0,
                      'BusyStatus' => 2);
        $attr = $content->getAllAttributes();
        foreach ($attr as $item) {
            $GLOBALS['backend']->logMessage("Sync4j for name=" . $item['name']
                                            . ", value=" . $item['value'] , __FILE__, __LINE__, PEAR_LOG_DEBUG);

            switch (String::upper($item['name'])) {
            case 'DTSTART':
                if (!empty($item['params']['VALUE']) &&
                    $item['params']['VALUE'] == 'DATE') {
                    $hash['AllDayEvent'] = 1;
                    $hash['Start'] = sprintf('%04d-%02d-%02d',
                                             $item['value']['year'],
                                             $item['value']['month'],
                                             $item['value']['mday']);
                    $start = mktime(0, 0, 0, $item['value']['month'], $item['value']['mday'], $item['value']['year']);
                } else {
                    $hash['AllDayEvent'] = 0;
                    $hash['Start'] = Horde_iCalendar::_exportDateTime($item['value']);
                    $start = $item['value'];
                }
                break;

            case 'DTEND':
                if (!empty($item['params']['VALUE']) &&
                    $item['params']['VALUE'] == "DATE") {
                    $hash['AllDayEvent'] = 1;
                    $hash['End'] = sprintf('%04d-%02d-%02d',
                                           $item['value']['year'],
                                           $item['value']['month'],
                                           $item['value']['mday']);
                    $end = mktime(0, 0, 0, $item['value']['month'], $item['value']['mday'], $item['value']['year']);
                } else {
                    $hash['AllDayEvent'] = 0;
                    $hash['End'] = Horde_iCalendar::_exportDateTime($item['value']);
                    $end = $item['value'];
                }
                break;

            case 'SUMMARY':
                $hash['Subject'] = $item['value'];
                break;

            case 'DESCRIPTION':
                $hash['Body'] = $item['value'];
                break;

            case 'LOCATION':
                $hash['Location'] = $item['value'];
                break;

            case 'CATEGORIES':
                $hash['Categories'] = $item['value'];
                break;

            case 'AALARM':
                $hash['ReminderSet'] = 1;
                $alarm = $item['value'];
                break;

            case 'STATUS':
                switch (String::upper($item['value'])) {
                case 'FREE':
                    $hash['BusyStatus'] = 0;
                    break;

                case 'TENTATIVE':
                    $hash['BusyStatus'] = 1;
                    break;

                case 'CONFIRMED':
                    $hash['BusyStatus'] = 2;
                    break;
                }
                break;

            case 'CLASS':
                switch (String::upper($item['value'])) {
                case 'PUBLIC':
                    $hash['Sensitivity'] = 0;
                    break;

                case 'PRIVATE':
                    $hash['Sensitivity'] = 2;
                    break;

                case 'CONFIDENTIAL':
                    $hash['Sensitivity'] = 3;
                    break;
                }
                break;

            case 'RRULE':
                // Parse the recurrence rule into keys and values.
                $rdata = array();
                $parts = explode(';', $item['value']);
                foreach ($parts as $part) {
                    list($key, $value) = explode('=', $part, 2);
                    $rdata[String::upper($key)] = $value;
                }

                if (!isset($rdata['FREQ'])) {
                    break;
                }

                // Always default the recurInterval to 1.
                $hash['Interval'] = isset($rdata['INTERVAL']) ? $rdata['INTERVAL'] : 1;
                $hash['IsRecurring'] = 1;

                switch (String::upper($rdata['FREQ'])) {
                case 'DAILY':
                    $hash['RecurrenceType'] = 0;
                    break;

                case 'WEEKLY':
                    $hash['RecurrenceType'] = 1;
                    if (isset($rdata['BYDAY'])) {
                        $maskdays = array('SU' => HORDE_DATE_MASK_SUNDAY,
                                          'MO' => HORDE_DATE_MASK_MONDAY,
                                          'TU' => HORDE_DATE_MASK_TUESDAY,
                                          'WE' => HORDE_DATE_MASK_WEDNESDAY,
                                          'TH' => HORDE_DATE_MASK_THURSDAY,
                                          'FR' => HORDE_DATE_MASK_FRIDAY,
                                          'SA' => HORDE_DATE_MASK_SATURDAY);
                        $days = explode(',', $rdata['BYDAY']);
                        $mask = 0;
                        foreach ($days as $day) {
                            $mask |= $maskdays[$day];
                        }
                        $hash['DayOfWeekMask'] = $mask;
                    }
                    break;

                case 'MONTHLY':
                    if (isset($rdata['BYDAY'])) {
                        $hash['RecurrenceType'] = 2;
                    } else {
                        $hash['RecurrenceType'] = 3;
                    }
                    break;

                case 'YEARLY':
                    $hash['RecurrenceType'] = 6;
                    break;
                }

                if (isset($rdata['UNTIL'])) {
                    $hash['NoEndDate'] = 0;
                    $hash['PatternEndDate'] = $rdata['UNTIL'];
                } else {
                    $hash['NoEndDate'] = 1;
                }
                break;
            }
        }

        if (empty($a['AllDayEvent']) && !empty($start) && !empty($end) && !empty($start) && $start!=$end) {
            $hash['Duration'] = ($end - $start) / 60;
            $GLOBALS['backend']->logMessage("duration set to=" . $hash['Duration']
            , __FILE__, __LINE__, PEAR_LOG_DEBUG);

        }
        if ($hash['ReminderSet'] && !empty($start) && !empty($alarm) && $start != $alarm) {
            $hash['ReminderMinutesBeforeStart'] = ($start - $alarm) / 60;
        }

        return SyncML_Device_sync4j::array2sif(
            $hash,
            '<?xml version="1.0"?><appointment>',
            '</appointment>');
    }

    function vtodo2sif($vcard)
    {
        $iCal = &new Horde_iCalendar();
        if (!$iCal->parsevCalendar($vcard)) {
            return PEAR::raiseError('There was an error importing the data.');
        }

        $components = $iCal->getComponents();

        switch (count($components)) {
        case 0:
            return PEAR::raiseError('No data was found');

        case 1:
            $content = $components[0];
            break;

        default:
            return PEAR::raiseError('Multiple components found; only one is supported.');
        }

        $hash['Complete'] = 'False';
        // Outlook's default for no due date
        $hash['DueDate'] = '45001231T230000Z';

        $attr = $content->getAllAttributes();
        foreach ($attr as $item) {
            switch ($item['name']) {
            case 'DUE':
                $hash['DueDate'] = Horde_iCalendar::_exportDateTime($item['value']);
                break;

            case 'SUMMARY':
                $hash['Subject'] = $item['value'];
                break;

            case 'DESCRIPTION':
                $hash['Body'] = $item['value'];
                break;

            case 'PRIORITY':
                if ($item['value'] == 1) {
                    $hash['Importance'] = 2;
                } elseif ($item['value'] == 5) {
                    $hash['Importance'] = 0;
                } else {
                    $hash['Importance'] = 1;
                }
                break;

            case 'COMPLETED':
                $hash['Complete'] = $item['value'] ? 'True' : 'False';
                break;
            case 'CATEGORIES':
                $hash['Categories'] = $item['value'];
                break;
            }
        }

        return SyncML_Device_sync4j::array2sif(
            $hash,
            '<?xml version="1.0"?><task>',
            '</task>');
    }

    /**
     * Sync4j as of Funambol Outlook connector 3.0.15 can't deal
     * with <![CDATA[ so omit it.
     */
    function useCdataTag()
    {
        return false;
    }

}
